package net.sf.corn.cps;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Modifier;
import java.net.JarURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

/**
 * This is the main class of this package. Resource scanner as it named, scans the java class paths
 * to find resources that are filtered out by the given filters to its scan methods. 
 * @author serhat.dirik
 */
public class CPScanner {

	/**
	 * Finds list of classpath roots
	 * @return list of class path root url's
	 */
	public static Set<URL> findRoots() {
		Set<URL> urls = new HashSet<URL>();
		ClassLoader loader = Thread.currentThread().getContextClassLoader();

		if (loader instanceof URLClassLoader) {
			URLClassLoader urlLoader = (URLClassLoader) loader;
			urls.addAll(Arrays.asList(urlLoader.getURLs()));
		} else {
			Enumeration<URL> urlEnum;
			try {
				urlEnum = loader.getResources("");
				while (urlEnum.hasMoreElements()) {
					urls.add(urlEnum.nextElement());
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		return urls;
	}
	
	/**
	 * Scans resources with in the classpath and applies the given list of filters 
	 * @param filters that need to be applied
	 * @return list of found resources
	 */
	public static List<URL> scanResources(CPResourceFilter... filters) {
		List<URL> resources = new ArrayList<URL>();
		Set<URL> roots = findRoots();
		filterResources(roots, filters);
		for (URL rootURL : roots) {
			resources.addAll(scan(rootURL, false, filters));
		}
		return resources;
	}

	/**
	 * Scans classes with in the classpath and applies the given list of filters 
	 * @param filters that need to be applied
	 * @return list of classes that found 
	 */
	public static List<Class<?>> scanClasses(CPResourceFilter... filters) {
		List<Class<?>> classList = new ArrayList<Class<?>>();

		Set<URL> roots = findRoots();
		filterResources(roots, filters);
		if (roots.size() == 0)
			return classList;
		List<URL> resources = null;
		ClassLoader loader = Thread.currentThread().getContextClassLoader();
		for (URL rootURL : roots) {
			resources = scan(rootURL, true, filters);
			for (URL resourceUrl : resources) {
				int extIndex = resourceUrl.toExternalForm().indexOf(".class");
				if (extIndex <= 0)
					continue;
				String className = null;
				if (isArchive(rootURL)) {
					className = resourceUrl.toExternalForm().substring(
							convertToJarUrl(rootURL).toExternalForm().length(),
							extIndex);
				} else {
					className = resourceUrl.toExternalForm().substring(
							rootURL.toExternalForm().length(), extIndex);
				}
				try {
					Class<?> clazz = loader.loadClass(className.replace('/',
							'.'));
					if (!clazz.isSynthetic()
							&& Modifier.isPublic(clazz.getModifiers())) {
						clazz.getName();
						clazz.getCanonicalName();
						if (acceptable(clazz, filters)) {
							classList.add(clazz);
						}
					}
				} catch (Exception  e) {
					// Runtime dependencies not complete to load this class,
					// just skip it
				}
			}
		}
		return classList;
	}

	private static void filterResources(Collection<?> resources,
			CPResourceFilter... filters) {
		Set<Object> removed = new HashSet<Object>();
		for (Object resource : resources) {
			if (!acceptable(resource, filters))
				removed.add(resource);
		}
		for (Object object : removed) {
			resources.remove(object);
		}
	}

	private static boolean acceptable(Object resource,
			CPResourceFilter... filters) {
		if (resource == null)
			return false;
		if (filters == null)
			return true;
		for (CPResourceFilter filter : filters) {
			if (filter.filterable(resource)) {
				if (!filter.accept(resource))
					return false;
			}
		}
		return true;
	}

	private static boolean isArchive(File f) {
		return f != null
				&& f.isFile()
				&& (f.getName().toLowerCase().endsWith("jar") || f.getName()
						.toLowerCase().endsWith("zip"));
	}

	private static boolean isArchive(URL url) {
		if (url.toString().startsWith("file")) {
			return url.toString().toLowerCase().endsWith("jar")
					|| url.toString().toLowerCase().endsWith("zip");
		} else {
			return url.toString().toLowerCase().startsWith("jar")
					|| url.toString().toLowerCase().startsWith("zip");
		}
	}

	private static boolean isClassUrl(URL url) {
		return url.toExternalForm().indexOf(".class") > 0;
	}

	private static URL convertToJarUrl(URL url) {
		URL jarUrl;
		try {
			jarUrl = new URL("jar:" + url + "!/");
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			jarUrl = url;
		}
		return jarUrl;
	}

	private static List<URL> scan(URL rootURL, boolean scan4Classes,
			CPResourceFilter... filters) {
		return scan(rootURL, rootURL, scan4Classes, filters);
	}

	private static List<URL> scan(URL rootURL, URL url, boolean scan4Classes,
			CPResourceFilter... filters) {

		List<URL> resources = new ArrayList<URL>();
		if (url.getFile() != null && url.getFile().length() > 0) {
			try {
				File file = new File(URLDecoder.decode(url.getFile(), "UTF-8"));
				if (file.isDirectory()) {
					// package! scan for childs
					for (File child : file.listFiles()) {
						resources.addAll(scan(rootURL, child.toURI().toURL(),
								scan4Classes, filters));
					}
				} else if (isArchive(file)) {// archive
					URL jarUrl = convertToJarUrl(file.toURI().toURL());
					if (acceptable(jarUrl, filters)) {
						JarURLConnection connection = (JarURLConnection) jarUrl
								.openConnection();
						JarFile jarFile = connection.getJarFile();
						if (acceptable(jarFile, filters)) {
							Enumeration<JarEntry> entries = jarFile.entries();
							while (entries.hasMoreElements()) {
								JarEntry entry = entries.nextElement();
								if (entry.isDirectory()) {
									continue;
								} else {
									URL entryUrl = new URL(jarUrl
											+ entry.toString());
									if (((scan4Classes && isClassUrl(entryUrl)) || (!scan4Classes && !isClassUrl(entryUrl)))
											&& acceptable(new RootedURL(jarUrl,entryUrl), filters)) {
										resources.add(entryUrl);
									}
								}
							}
						}
					}
				} else {
					// Resource under a package
					if ( ( (scan4Classes && isClassUrl(file.toURI().toURL()))
							|| (!scan4Classes && !isClassUrl(file.toURI()
									.toURL())) )
							&& acceptable(new RootedURL(rootURL, url), filters)) {
						resources.add(file.toURI().toURL());
					}
				}
			} catch (IOException e) {
				// TODO Handle this
				// We don't expect to trapped here
				e.printStackTrace();
			}
		}
		return resources;
	}

}
